//////////////////////////////////////////////////////////////////////////////
//
// 		  Copyright (C) 1996,1997  Matthew Doar  doar@pobox.com
// 
// Permission to use, copy, and distribute this software and its documentation 
// for any purpose with or without fee is hereby granted, provided that the 
// above copyright notice appears with all copies and that both that copyright 
// notice and this permission notice appear in supporting documentation. 
// 
// Permission to modify the software is granted, but not the right to 
// distribute the modified code. Modifications are to be distributed as 
// patches to the released version. 
// 
// This software is provided "as is" without express or implied warranty. 
//
//////////////////////////////////////////////////////////////////////////////

// tiers_pnni.cc

#include <iostream.h>
#include <limits.h>	// for ULONG_MAX

#ifndef _TIERS_HH
#include "tiers.hh"
#endif

extern "C" 
{
  long random(void);
}

#ifndef _TIERS_PNNI_HH
#include "tiers_pnni.hh"
#endif



////////////////////////////////////////////////////////////////////////
// PnniModel::PnniModel
//
// Constructor for top level PNNI class
////////////////////////////////////////////////////////////////////////
PnniModel::PnniModel(void)
{
}



////////////////////////////////////////////////////////////////////////
// PnniModel::~PnniModel
//
// Destructor for top level class
////////////////////////////////////////////////////////////////////////
PnniModel::~PnniModel(void)
{
}


////////////////////////////////////////////////////////////////////////
// PnniModel::ConnectMans
//
// Connect each MAN to a unique WAN node
// Add the WAN connection point coordinates to all the nodes in the MAN
// Also add the internetwork redundancy from MANs to WAN
////////////////////////////////////////////////////////////////////////
bool PnniModel::ConnectMans()
{
  bool ret = true;

  unsigned long int NumWanNodes = WanList[0].NumNodes();
  unsigned long int MaxNumWanEdges = WanList[0].MaxNumEdges();
  Edge *wanedges = WanList[0].newedges;	// an alias
  Node *wannodes = WanList[0].nodes;	// an alias

  // Step through each MAN, connecting the MAN to the WAN and vice versa
  for (unsigned long int network = 0; 
	   network < WanList[0].NumNetworks();
	   network++)
	{
	  // Select a WAN node as an attachment point for a MAN
	  // Duplicates are NOT permitted. 
	  unsigned long int wangateway = network;	//random() % NumWanNodes;

	  unsigned long int NumManNodes = ManList[network].NumNodes();
	  unsigned long int MaxNumManEdges = ManList[network].MaxNumEdges();

	  // Choose the index of the first node to be connected
	  unsigned long int gateway = random() % NumManNodes;

	  if (!AddInternetworkEdge(ManList, WanList,
							   gateway, wangateway, 
							   MaxNumManEdges, network)) 
		{
		  cerr << "ConnectMans:: Failed to add internetwork edge to MAN and WAN" << endl;
		  return false;
		}

	  // Update the coordinates of all the nodes in the MAN
	  if (!RepositionNetwork(&(ManList[network]), 
							 &(WanList[0].nodes[wangateway]),	
							   MAN))
		{
		  cerr << "ConnectMans:: Failed to reposition MAN" << endl;
		  return false;
		}


	  // Add edges for internetwork redundancy
	  // Add an edge from the MAN to the closest nodes to the original 
	  // wangateway. Increase the probability that once a node has been 
	  // chosen as a connection point in a MAN to the WAN, that it will be 
	  // chosen again, with p = 1/2.
  	  unsigned long int R = RMW-1;
	  while (R)
		{
		  // Choose the (possibly new) gateway if != 0
		  if ((random() % 2))
			{
			  gateway = random() % NumManNodes;
			}

		  // Find the new wangateway - node closest to current wangateway
		  // by  non INACTIVE edge.
		  // Could have used a full connectivity array here if it still existed
		  // The closest node cannot be self since no edge exists to self
		  unsigned long int NumWanEdges = WanList[0].NumEdges(wangateway);
		  unsigned long int EdgeClosestTo = NumWanEdges;
		  unsigned long int MinimumCost = ULONG_MAX;

		  for (unsigned long int i = 0; 
			   i < NumWanEdges; 
			   i++)
			{
			  // This explicitly uses delay for the metric
			  unsigned long int index = wangateway*MaxNumWanEdges+i;
			  unsigned long int cost = wanedges[index].delay;

			  // Make sure not to use the MAN edge (always closest to the WAN)
			  if ((cost < MinimumCost) && 
				  (wanedges[index].state != INACTIVE) &&
				  (wanedges[index].state < Model::INTERNETWORK))
				{
				  MinimumCost = cost;
				  EdgeClosestTo = i;
				}
			}
		  // There may only be one node in the WAN
		  if ((MinimumCost == ULONG_MAX) && (WanList[0].NumNodes() > 1))
			{ 
			  cerr << "ConnectMans:: WAN edge list corrupt" << endl;
			  return false;
			}

		  // Find the correct WAN index for ClosestTo.end
		  unsigned long int sought_node_num = 
			wanedges[wangateway*MaxNumWanEdges+EdgeClosestTo].end;
		  wangateway = 0;
		  while(wannodes[wangateway].number != 
				sought_node_num) 
			{
			  wangateway++;
			}
		  if (wangateway == NumWanNodes)
			{ 
			  cerr << "ConnectMans:: WAN node list corrupt" << endl;
			  return false;
			}

		  if (!AddInternetworkEdge(ManList, WanList,
								   gateway, wangateway, 
								   MaxNumManEdges, network,
								   Model::RED_INTERNETWORK)) 
			 {
			   cerr << "ConnectMans:: Failed to add redundant internetwork edge to MAN and WAN" << endl;
			   return false;
			 }

		  
		  // And decrement the number of redundant internetwork edges 
		  // still to be added
		  R--;

		}// while R
	} // network

  return ret;
}


////////////////////////////////////////////////////////////////////////
// PnniModel:OutputParams
//
// Output all the parameters and details needed to reconstruct the model
////////////////////////////////////////////////////////////////////////
bool
PnniModel::OutputParams()
{
  bool ret = true;
  
  if (TIERS_VERBOSE)
	{
	  // output the parameters used to create this network
	  cout << "#" << endl 
		   << "# This model is a PNNI Model" << endl 
		   << "#" << endl
		   << "# Parameters for Network Model\tModelId: " << ModelId << endl 
		   << "# ============================ " << endl 
		   << "#" << endl
		   << "# Value\tParameter" << endl
		   << "# -----\t---------" << endl 
		   << "#" << endl 
		   << "# " << NW << "\tNW: number of WANs" << endl
		   << "# " << NM << "\tNM: number of MANs per WAN" << endl
		   << "# " << NL << "\tNL: number of LANs per MAN" << endl
		   << "# " << SW << "\tSW: number of nodes per WAN (equal to NM)" << endl
		   << "# " << SM << "\tSM: number of nodes per MAN" << endl
		   << "# " << SL << "\tSL: number of nodes per LAN" << endl
		   << "# " << RW << "\tRW: intranetwork redundancy for WAN" << endl
		   << "# " << RM << "\tRM: intranetwork redundancy for MANs" << endl
		   << "# " << RL << "\tRL: intranetwork redundancy for LANs" << endl
		   << "# " << RMW << "\tRMW: internetwork redundancy for MAN to WAN" << endl
		   << "# " << RLM << "\tRLM: internetwork redundancy for LAN to MAN" << endl
		   << "#" << endl
		   << "# " << GRID << "\t Grid size (square)" << endl
		   << "# " << WAN_SCALE << "\t Grid unit for WAN" << endl
		   << "# " << MAN_SCALE << "\t Grid unit for MAN" << endl
		   << "# " << LAN_SCALE << "\t Grid unit for LAN" << endl
		   << "#" << endl
		   << "# " << PROXIMITY_TEST_WAN << "\t Proximity test for WAN (0 is inactive)" << endl 
		   << "# " << PROXIMITY_TEST_MAN << "\t Proximity test for MAN (0 is inactive)" << endl 
		   << "# " << PROXIMITY_TEST_LAN << "\t Proximity test for LAN (0 is inactive)" << endl << "#" << endl
		   << "# " << REMOVE_DUP_EDGES << "\t Edge list is directed (0) or undirected (1)" << endl;

	  unsigned long int TotalNumNodes = SW + (NM*SM) + (NM*NL*SL);
	  cout << "# Total Number of Nodes (derived) : " << TotalNumNodes << endl;

	  // output the Cost Table used for this network
	  cout << "#" << endl << "# Edge Metrics" << endl 
		   << "# ------------" << endl
		   << "#" << endl << "# From\tTo\tBand-\tFixed\tDelay/Unit" << endl
		   << "# Type\tType\twidth\tDelay\tDistance" << endl
		   << "# ------------------------------------------" << endl;

	  for (unsigned int i = 0; i < 3; i++)
		{
		  for (unsigned int j = 0; j < 3; j++)
			{
			  unsigned int index = i*3+j;
			  cout 
				<< "# " << i << "\t" << j << "\t"
				<< theCostTable[index].bw << "\t"
				<< theCostTable[index].d1 << "\t"
				<< theCostTable[index].d2 << "\t"
				<< endl;
			}
		}

	} // ifdef

  return ret;
}




////////////////////////////////////////////////////////////////////////
// PnniModel:OutputNodes
//
// Output all the nodes in the graph
////////////////////////////////////////////////////////////////////////
bool
PnniModel::OutputNodes()
{
  bool ret = true;
  unsigned long int i = 0;
  unsigned long int j = 0;
  
  if (TIERS_VERBOSE)
	{
	  cout << "#" << endl
		   << "# NODES" << endl
		   << "# =====" << endl 
		   << "#" << endl
		   << "# Node\tX\tY\tType (0 = WAN, 1 = MAN, 2 = LAN)" << endl
		   << "# ------------------------------------" << endl;

	}

  // WAN - subsumed into MAN nodes
  
  // MANs - already scaled when repositioning took place
  unsigned long int NumMans = WanList[0].NumNetworks();
  for (i = 0; i < NumMans; i++)
	{
	  if (TIERS_VERBOSE)
		{
		  cout << "#" << endl
			   << "# Level 1 Peer Group " << i << endl
			   << "#" << endl;
		}
	  cout << ((WanList[0]).nodes[i]).number << "\t" 
		   << (((WanList[0]).nodes[i]).x)*WAN_SCALE << "\t" 
		   << (((WanList[0]).nodes[i]).y)*WAN_SCALE << "\t" 
		   << ((WanList[0]).nodes[i]).type << endl;
	  for (j = 0; j < ManList[i].NumNodes(); j++)
		{
		  cout << ((ManList[i]).nodes[j]).number << "\t" 
			   << (((ManList[i]).nodes[j]).x) << "\t" 
			   << (((ManList[i]).nodes[j]).y) << "\t" 
			   << ((ManList[i]).nodes[j]).type << endl;
		}
	}

  // LANs - already scaled when repositioning took place
  // Count the total number of LANs
  unsigned long int NumLans = 0;
  for (unsigned long int man = 0; man < NumMans; man++)
	{
	  NumLans += ManList[man].NumNetworks();
	}

  bool newLan;	// set when the first node in a LAN is output
  // NOTE: Not all these LANs are valid entries
  for (i = 0; i < NumMans*NumLans; i++)
	{
	  newLan = true;
	  for (j = 0; j < LanList[i].NumNodes(); j++)
		{
		  if (TIERS_VERBOSE && newLan)
			{
			  cout << "#" << endl
				   << "# Level 2 Peer Group " << i << endl
				   << "#" << endl;
			  newLan = false;	// don't output the comment more than once
			}
		  cout << ((LanList[i]).nodes[j]).number << "\t" 
			<< (((LanList[i]).nodes[j]).x) << "\t" 
			<< (((LanList[i]).nodes[j]).y) << "\t" 
			<< ((LanList[i]).nodes[j]).type << endl;
		}
	}
  
  return ret;
}



////////////////////////////////////////////////////////////////////////
// PnniModel::OutputEdges
//
// Output all the active edges in the graph
////////////////////////////////////////////////////////////////////////
bool
PnniModel::OutputEdges()
{
  bool ret = true;
  unsigned long int network = 0;
  unsigned long int from = 0;
  
  if (TIERS_VERBOSE)
	{
	  cout << "#" << endl
		   << "# EDGES" << endl
		   << "# =====" << endl 
		   << "#" << endl
		   << "# From\t To\tDelay\tBand-\tFrom\tTo\tState (1:active, 2:redundant" << endl
		   << "# Node\tNode\t     \twidth\tType\tType\t3:internetwork, 4:red.inter.)"
		   << endl
		   << "# ---------------------------------------------------" << endl;
	}


  // WANs
  for (network = 0; network < 1 /* NW should equal 1 */; network++)
	{
	  if (TIERS_DEBUG)
		{
		  cout << "#" << endl 
			   << "# Inter level 1 peer groups"  << endl 
			   << "#" << endl;
		}

	  for (from = 0; from < WanList[network].NumNodes(); from++)
		{
		  unsigned long int MaxNumEdges = WanList[network].MaxNumEdges();
		  unsigned long int NumEdges = WanList[network].NumEdges(from);
		  Edge	*edges = WanList[network].newedges;
		  for (unsigned long int to = 0; to < NumEdges; to++)
			{
			  unsigned long int index = from*MaxNumEdges+to;
			  Node fromNode = WanList[network].nodes[from];
			  Node toNode = WanList[network].nodes[to];

			  if (!(edges[index].state == Model::INACTIVE) && 
				  WanList[network].test_dup_edges(edges[index].start, 
												  edges[index].end))
				{
				  cout << edges[index].start << "\t"
					   << edges[index].end << "\t"
					   << edges[index].delay << "\t"
					   << edges[index].bw << "\t"
					   << fromNode.type << "\t";

				  if (edges[index].state < Model::INTERNETWORK)
					{
					  cout << toNode.type;
					}
				  else
					{
					  cout << MAN;
					}

				  cout << "\t" << edges[index].state
					   << endl;
				}
			} // to
		} // from
	} // network

  // MANs
  for (network = 0; network < WanList[0].NumNetworks(); network++)
	{
	  if (TIERS_DEBUG)
		{
		  cout << "#" << endl 
			   << "# Intra level 1 peer group " << network << endl 
			   << "#" << endl;
		}

	  for (from = 0; from < ManList[network].NumNodes(); from++)
		{
		  unsigned long int MaxNumEdges = ManList[network].MaxNumEdges();
		  unsigned long int NumEdges = ManList[network].NumEdges(from);
		  Edge	*edges = ManList[network].newedges;
		  for (unsigned long int to = 0; to < NumEdges; to++)
			{
			  unsigned long int index = from*MaxNumEdges+to;
			  Node fromNode = ManList[network].nodes[from];
			  // a useful alias, only valid if `to' is in the current network
			  Node toNode = ManList[network].nodes[to];

			  if (!(edges[index].state == Model::INACTIVE) &&
				  ManList[network].test_dup_edges(edges[index].start,
												  edges[index].end))
				{
				  cout << edges[index].start << "\t"
					   << edges[index].end << "\t"
					   << edges[index].delay << "\t"
					   << edges[index].bw << "\t"	
					   << fromNode.type << "\t";

				  // Problem: how to find the type of the destination node
				  // if the edge is not in the current network.
				  // Partial solution: if the edges[index].state is ACTIVE or 
				  // REDUNDANT, the node is in the current network.
				  // This resulted in bug noted 3/3/97 in CHANGES
				  // Better partial solution - test if the destination node 
				  // number is in the range for the WAN. DONE
				  // Best solution: create a global node list as is done in
				  // tiers_output_gnuplot.cc
				  if (edges[index].state < Model::INTERNETWORK)
					{
					  cout << toNode.type;
					}
				  else
					{
					  if (edges[index].end < SW)
						{
						  cout << WAN;	// relies on unique node numbers
						}
					  else
						{
						  cout << LAN;
						}
					}

				  cout << "\t" << edges[index].state
					   << endl;
				}
			} // to
		} // from
	} // network

  // LANs
  // LANs are modelled as a star, with the edges in edges
  // Edges for internetwork redundancy are at the end of edges
  network = 0; // the index of each LAN
  for (unsigned long int supernetwork = 0; 
	   supernetwork < WanList[0].NumNetworks(); 
	   supernetwork++)
	{
	  unsigned long int UpperLimit = network + 
		ManList[supernetwork].NumNetworks();
	  for (network = network; network < UpperLimit; network++)
		{
		  if (TIERS_DEBUG)
			{
			  cout << "#" << endl 
				   << "# Intra level 2 peer group " << network << endl 
				   << "#" << endl;
			}

		  // edge (0,0) is undefined
		  for (unsigned long int to = 1; 
			   to < (LanList[network].NumEdges(0)+1); 
			   to++)
			{
			  Edge	*edges = LanList[network].edges;	// newedges is 0
			  Node fromNode = LanList[network].nodes[0];
			  if (edges[to].state != Model::INACTIVE)
				{
				  cout << edges[to].start << "\t"
					   << edges[to].end << "\t"
					   << edges[to].delay << "\t"
					   << edges[to].bw << "\t"
					   << fromNode.type << "\t";

				  if (edges[to].state < Model::INTERNETWORK)
					{
					  cout << LAN;
					}
				  else
					{
					  cout << MAN;
					}

				  cout << "\t" << edges[to].state
					   << endl;

				  // generate the symmetrical edge for the purposes of output
				  if (!REMOVE_DUP_EDGES)
					{
					  cout << edges[to].end << "\t"
						   << edges[to].start << "\t"
						   << edges[to].delay << "\t"
						   << edges[to].bw << "\t";

					  if (edges[to].state < Model::INTERNETWORK)
						{
						  cout << LAN;
						}
					  else
						{
						  cout << MAN;
						}

					  cout << "\t" << fromNode.type
						   << "\t" << edges[to].state
						   << endl;
					}
				} // to
			} // network
		} // supernetwork
	}
  cout << "# End of Model" << endl;
  
  return ret;
}


// GNUPLOT functions come below here


////////////////////////////////////////////////////////////////////////
// PnniModel:CreateGlobalNodeList
//
// Create a global list of all the nodes for use when outputing the edges
// Output labels for in the model, in a format suitable for gnuplot
////////////////////////////////////////////////////////////////////////
bool
PnniModel::CreateGlobalNodeList(Node* &GlobalNodeList)
{
  bool ret = true;
  unsigned long int i = 0;
  unsigned long int j = 0;

  // The total number of nodes
  unsigned long int NumNodes = WanList[0].NumNodes(); 
  unsigned long int NumLans = 0;
  unsigned long int CurrentNode = 0;

  // Count the total number of nodes to know how large an array to allocate
  unsigned long int lanindex = 0;
  for (unsigned long int man = 0; man < WanList[0].NumNetworks(); man++)
	{
	  NumNodes += ManList[man].NumNodes();
	  NumLans += ManList[man].NumNetworks();

	  unsigned long int UpperLimit = lanindex + ManList[man].NumNetworks();
	  for (lanindex = lanindex; lanindex < UpperLimit; lanindex++)
	    {
	      NumNodes += LanList[lanindex].NumNodes();
	    }
	}

  GlobalNodeList = new Node[NumNodes];
  if (!GlobalNodeList) 
	{
	  cerr << "Tiers:: failed to allocate space for global node list" << endl;
	  ret = false;
	}


  if (TIERS_LABEL_WAN || TIERS_LABEL_MAN || TIERS_LABEL_LAN)
	{
	  cout << "#" << endl
		   << "# NODES" << endl
		   << "# =====" << endl 
		   << "#" << endl
		   << "# Node Label\tlabelX\tlabelY" << endl
		   << "# ------------------------------------" << endl;
	}

  //
  // Copy each of the node details into nodes
  // and optionally print out the coordinates and a label (the node number)
  //

  // The offset for the label from the node, in the WAN units
  unsigned long int offsetX = 1;
  unsigned long int offsetY = 1;

  // WAN
  for (i = 0; i < 1 /* NW should equal 1 */; i++)
	{
	  for (j = 0; j < WanList[i].NumNodes(); j++)
		{
		  GlobalNodeList[CurrentNode] = WanList[i].nodes[j];

		  if (TIERS_LABEL_WAN)
		    {
			  cout << "set label \"" << CurrentNode << "\" at "
				   << (((WanList[i]).nodes[j]).x + offsetX)*WAN_SCALE << "," 
				   << (((WanList[i]).nodes[j]).y + offsetY)*WAN_SCALE 
				   << " font \"Times,8\""
				   << endl;
			}
		  
		  CurrentNode++;
		}
	}
  
  // MANs
  unsigned long int NumMans = WanList[0].NumNetworks();
  for (i = 0; i < NumMans; i++)
	{
	  for (j = 0; j < ManList[i].NumNodes(); j++)
		{
		  GlobalNodeList[CurrentNode] = ManList[i].nodes[j];

		  if (TIERS_LABEL_MAN)
		    {
			  cout << "set label \"" << CurrentNode << "\" at "
				   << (((ManList[i]).nodes[j]).x) + offsetX << "," 
				   << (((ManList[i]).nodes[j]).y) + offsetY 
				   << " font \"Times,8\""
				   << endl;
			}

		  CurrentNode++;
		}
	}

  // LANs
  for (i = 0; i < NumLans; i++)
	{
	  for (j = 0; j < LanList[i].NumNodes(); j++)
		{
		  GlobalNodeList[CurrentNode] = LanList[i].nodes[j];

		  if (TIERS_LABEL_LAN)
		    {
			  cout << "set label \"" << CurrentNode << "\" at "
				   << (((LanList[i]).nodes[j]).x) + offsetX << "," 
				   << (((LanList[i]).nodes[j]).y) + offsetY 
				   << " font \"Times,8\""
				   << endl;
			}

		  CurrentNode++;
		}
	}
  

  return ret;
}



////////////////////////////////////////////////////////////////////////
// PnniModel::OutputEdgesGnuplot
//
// Output all the active edges in the graph
// Basically identical to OutputEdges but kept separate for future flexibility
////////////////////////////////////////////////////////////////////////
bool
PnniModel::OutputEdgesGnuplot()
{
  bool ret = true;
  unsigned long int network = 0;
  unsigned long int from = 0;

  Node *GlobalNodeList = 0;
  // Create a list of all the nodes for use when outputing the edges
  // As a side-effect, it may print out labels for each of the nodes
  if (!CreateGlobalNodeList(GlobalNodeList))
    {
	  cerr << "Tiers:: create global node list output error" << endl;
	  ret = false;
    }

  // Write the plot statement into the script, then add the data to be plotted
  // Check that MAN/LAN nodes exist before printing each plot out
  cout << "plot '-' title 'Backbone' with linespoints 3";

  if (ManList[0].NumNodes() > 1)
	cout << ", '-' title 'Level 1' with linespoints 2";

  if (LanList[0].NumNodes() > 1)
	cout << ", '-' title 'Level 2' with linespoints 1";
  // And end the line
  cout<< endl;


  if (TIERS_VERBOSE)
	{
	  cout << "#" << endl
		   << "# EDGES" << endl
		   << "# =====" << endl 
		   << "#" << endl
		   << "# From\t(x1, y1)" << endl
		   << "# To\t(x2, y2)" << endl
		   << "# -------------" << endl;
	}

  // Repositioning took care of scaling the coordinates appropriately

  // WANs
  for (network = 0; network < 1 /* NW should equal 1 */; network++)
	{
	  if (TIERS_DEBUG)
		{
		  cout << "#" << endl 
			   << "# Inter level 1 peer groups"  << endl 
			   << "#" << endl;
		}

	  for (from = 0; from < WanList[network].NumNodes(); from++)
		{
		  unsigned long int MaxNumEdges = WanList[network].MaxNumEdges();
		  unsigned long int NumEdges = WanList[network].NumEdges(from);
		  Edge	*edges = WanList[network].newedges;
		  for (unsigned long int to = 0; to < NumEdges; to++)
			{
			  unsigned long int index = from*MaxNumEdges+to;
			  unsigned long int start = edges[index].start;
			  unsigned long int end = edges[index].end;

			  unsigned long int start_scale = WAN_SCALE;
			  if (GlobalNodeList[start].type > WAN)
				{
				  start_scale = 1;	// MAN positions have already been scaled
				}
			  
			  unsigned long int end_scale = WAN_SCALE;
			  if (GlobalNodeList[end].type > WAN)
				{
				  end_scale = 1;	// MAN positions have already been scaled
				}
			  
			  if ((edges[index].state != Model::INACTIVE) && 
				  (WanList[network].test_dup_edges(edges[index].start, 
												   edges[index].end)) )
				{
				  cout << GlobalNodeList[start].x * start_scale << "\t"
					   << GlobalNodeList[start].y * start_scale << endl
					   << GlobalNodeList[end].x * end_scale << "\t"
					   << GlobalNodeList[end].y * end_scale 
					   << endl << endl;
				}
			} // to
		} // from
	} // network

  // End the inline dataset
  cout << "e" << endl;

  // MANs
  for (network = 0; network < WanList[0].NumNetworks(); network++)
	{
	  if (TIERS_DEBUG)
		{
		  cout << "#" << endl 
			   << "# Intra level 1 peer group " << network << endl 
			   << "#" << endl;
		}

	  for (from = 0; from < ManList[network].NumNodes(); from++)
		{
		  unsigned long int MaxNumEdges = ManList[network].MaxNumEdges();
		  unsigned long int NumEdges = ManList[network].NumEdges(from);
		  Edge	*edges = ManList[network].newedges;
		  for (unsigned long int to = 0; to < NumEdges; to++)
			{
			  unsigned long int index = from*MaxNumEdges+to;
			  unsigned long int start = edges[index].start;
			  unsigned long int end = edges[index].end;

			  unsigned long int start_scale = 1;
			  if (GlobalNodeList[start].type == WAN)
				{
				  // WAN positions have not already been scaled
				  start_scale = WAN_SCALE;	
				}
			  
			  unsigned long int end_scale = 1;
			  if (GlobalNodeList[end].type == WAN)
				{
				  // WAN positions have not already been scaled
				  end_scale = WAN_SCALE;	
				}
			  
			  if ((edges[index].state != Model::INACTIVE) &&
				  (ManList[network].test_dup_edges(edges[index].start,
												   edges[index].end)))
				{
				  cout << GlobalNodeList[start].x * start_scale << "\t"
					   << GlobalNodeList[start].y * start_scale << endl
					   << GlobalNodeList[end].x * end_scale << "\t"
					   << GlobalNodeList[end].y * end_scale
					   << endl << endl;
				}
			} // to
		} // from
	} // network

  // End the inline dataset
  if (ManList[0].NumNodes() > 1)
	cout << "e" << endl;

  // LANs
  // LANs are modelled as a star, with the edges in edges
  // Edges for internetwork redundancy are at the end of edges
  network = 0; // the index of each LAN
  for (unsigned long int supernetwork = 0; 
	   supernetwork < WanList[0].NumNetworks(); 
	   supernetwork++)
	{
	  if (TIERS_DEBUG)
		{
		  cout << "#" << endl 
			   << "# Intra level 2 peer group " << network << endl 
			   << "#" << endl;
		}


	  unsigned long int UpperLimit = network + 
		ManList[supernetwork].NumNetworks();
	  for (network = network; network < UpperLimit; network++)
		{
		  // edge (0,0) is undefined
		  for (unsigned long int to = 1; 
			   to < (LanList[network].NumEdges(0)+1); 
			   to++)
			{
			  Edge	*edges = LanList[network].edges;	// newedges is 0
			  unsigned long int start = edges[to].start;
			  unsigned long int end = edges[to].end;
			  
			  // No scaling needed since LANs only ever connect to MANs and
			  // both have already been scaled when repositioning them

			  if (edges[to].state != Model::INACTIVE)
				{
				  cout << GlobalNodeList[start].x << "\t"
					   << GlobalNodeList[start].y << endl
					   << GlobalNodeList[end].x << "\t"
					   << GlobalNodeList[end].y 
					   << endl << endl;

				  // generate the symmetrical edge for the purposes of output
				  if (!REMOVE_DUP_EDGES)
					{
					  cout << GlobalNodeList[end].x << "\t"
						   << GlobalNodeList[end].y << endl
						   << GlobalNodeList[start].x << "\t"
						   << GlobalNodeList[start].y 
						   << endl << endl;

					}
				} // to
			} // network
		} // supernetwork
	}

  // End the inline dataset
  if (LanList[0].NumNodes() > 1)
	cout << "e" << endl;

  return ret;
}

  
// end of file

